"""Test plotting."""

import os

import pytest
import sympy

import symfem
import symfem.plotting

folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../_temp")
os.system(f"mkdir {folder}")


def compile_tex(filename):
    with open(os.path.join(folder, filename)) as f:
        content = f.read()
    filename2 = filename.split(".")[0] + "-with-preamble.tex"
    with open(os.path.join(folder, filename2), "w") as f:
        f.write("\\documentclass{standalone}\n")
        f.write("\\usepackage{tikz}\n")
        f.write("\\begin{document}\n")
        f.write(content)
        f.write("\\end{document}\n")
    assert os.system(f"cd {folder} && pdflatex -halt-on-error {filename2} > /dev/null") == 0


def test_plot_line():
    p = symfem.plotting.Picture()
    p.add_line((sympy.Integer(0), sympy.Integer(0)),
               (sympy.Integer(1), sympy.Integer(1)), "blue")
    for ext in ["svg", "png", "tex"]:
        p.save(os.path.join(folder, f"test_plot_line.{ext}"))
    compile_tex("test_plot_line.tex")


def test_plot_arrow():
    p = symfem.plotting.Picture()
    p.add_arrow((sympy.Integer(0), sympy.Integer(0)),
                (sympy.Integer(1), sympy.Integer(1)), "orange")
    for ext in ["svg", "png", "tex"]:
        p.save(os.path.join(folder, f"test_plot_arrow.{ext}"))
    compile_tex("test_plot_arrow.tex")


def test_plot_ncircle():
    p = symfem.plotting.Picture()
    p.add_ncircle((sympy.Integer(0), sympy.Integer(0)), 5, "green")
    for ext in ["svg", "png", "tex"]:
        p.save(os.path.join(folder, f"test_plot_ncircle.{ext}"))
    compile_tex("test_plot_ncircle.tex")


def test_plot_fill():
    p = symfem.plotting.Picture()
    p.add_fill((
        (sympy.Integer(0), sympy.Integer(0)), (sympy.Integer(1), sympy.Integer(1)),
        (sympy.Integer(1), sympy.Rational(1, 2)), (sympy.Rational(1, 2), sympy.Integer(0))
    ), "purple", 0.5)
    for ext in ["svg", "png", "tex"]:
        p.save(os.path.join(folder, f"test_plot_fill.{ext}"))
    compile_tex("test_plot_fill.tex")


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
    "tetrahedron", "hexahedron", "prism", "pyramid"
])
def test_z_ordering(reference):
    r = symfem.create_reference(reference)

    z = []
    for i in r.z_ordered_entities():
        z += i

    assert len(z) == len(set(z))

    for i in range(4):
        assert len([e for e in z if e[0] == i]) == r.sub_entity_count(i)


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
    "tetrahedron", "hexahedron", "prism", "pyramid"
])
@pytest.mark.parametrize("degree", [0, 2, 5])
def test_dof_diagrams_lagrange(reference, degree):
    e = symfem.create_element(reference, "P", degree)
    for ext in ["svg", "png", "tex"]:
        e.plot_dof_diagram(os.path.join(
            folder, f"test_dof_diagrams_lagrange-{reference}-{degree}.{ext}"))
    compile_tex(f"test_dof_diagrams_lagrange-{reference}-{degree}.tex")


@pytest.mark.parametrize("reference", [
    "triangle", "quadrilateral",
    "tetrahedron", "hexahedron",
])
@pytest.mark.parametrize("degree", [1, 2])
def test_dof_diagrams_raviart_thomas(reference, degree):
    if reference in ["quadrilateral", "hexahedron"]:
        e = symfem.create_element(reference, "Qdiv", degree)
    else:
        e = symfem.create_element(reference, "RT", degree)
    for ext in ["svg", "png", "tex"]:
        e.plot_dof_diagram(os.path.join(
            folder, f"test_dof_diagrams_raviart_thomas-{reference}-{degree}.{ext}"))
    compile_tex(f"test_dof_diagrams_raviart_thomas-{reference}-{degree}.tex")


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
])
@pytest.mark.parametrize("degree", [0, 1, 2])
def test_function_plots_lagrange(reference, degree):
    e = symfem.create_element(reference, "P", degree)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_lagrange-{reference}-{degree}.{ext}"))
    compile_tex(f"test_function_plots_lagrange-{reference}-{degree}.tex")


@pytest.mark.parametrize("reference", [
    "triangle", "quadrilateral",
    "tetrahedron", "hexahedron",
])
@pytest.mark.parametrize("degree", [1])
def test_function_plots_raviart_thomas(reference, degree):
    if reference in ["quadrilateral", "hexahedron"]:
        e = symfem.create_element(reference, "Qdiv", degree)
    else:
        e = symfem.create_element(reference, "RT", degree)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_raviart_thomas-{reference}-{degree}.{ext}"))
    compile_tex(f"test_function_plots_raviart_thomas-{reference}-{degree}.tex")


@pytest.mark.parametrize("reference", [
    "triangle", "quadrilateral",
])
def test_function_plots_piecewise_scalar(reference):
    e = symfem.create_element(reference, "P1-iso-P2", 1)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_piecewise_scalar-{reference}.{ext}"))
    compile_tex(f"test_function_plots_piecewise_scalar-{reference}.tex")


@pytest.mark.parametrize("reference", [
    "triangle", "tetrahedron",
])
def test_function_plots_piecewise_vector(reference):
    e = symfem.create_element(reference, "Guzman-Neilan", 1)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_piecewise_vector-{reference}.{ext}"))
    compile_tex(f"test_function_plots_piecewise_vector-{reference}.tex")


@pytest.mark.parametrize("n", [4, 6])
def test_function_plots_dual(n):
    e = symfem.create_element(f"dual polygon({n})", "dual", 1)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_dual-{n}.{ext}"))
    compile_tex(f"test_function_plots_dual-{n}.tex")


@pytest.mark.parametrize("n", [4, 6])
def test_function_plots_bc(n):
    e = symfem.create_element(f"dual polygon({n})", "BC", 1)
    for ext in ["svg", "png", "tex"]:
        e.plot_basis_function(0, os.path.join(
            folder, f"test_function_plots_bc-{n}.{ext}"))
    compile_tex(f"test_function_plots_bc-{n}.tex")


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
    "tetrahedron", "hexahedron", "prism", "pyramid",
    "dual polygon(6)"
])
def test_plot_reference(reference):
    r = symfem.create_reference(reference)
    rname = reference.replace("(", "").replace(")", "").replace(" ", "_")
    for ext in ["svg", "png", "tex"]:
        r.plot_entity_diagrams(os.path.join(
            folder, f"test_plot_reference-{rname}.{ext}"))
    compile_tex(f"test_plot_reference-{rname}.tex")


def test_metadata():
    img = symfem.plotting.Picture(svg_metadata=(
        "<metadata id='license'>\n"
        " <rdf:RDF xmlns:rdf='http://www.w3.org/1999/02/22-rdf-syntax-ns#' "
        "xmlns:dc='http://purl.org/dc/elements/1.1/' "
        "xmlns:cc='http://web.resource.org/cc/'>\n"
        "   <cc:Work rdf:about=''>\n"
        "     <dc:title>Title</dc:title>\n"
        "     <dc:date>1970-01-01</dc:date>\n"
        "     <dc:creator>\n"
        "       <cc:Agent><dc:title>Symfem</dc:title></cc:Agent>\n"
        "     </dc:creator>\n"
        "     <dc:description>See document description</dc:description>\n"
        "     <cc:license rdf:resource='http://creativecommons.org/licenses/by/4.0/'/>\n"
        "     <dc:format>image/svg+xml</dc:format>\n"
        "     <dc:type rdf:resource='http://purl.org/dc/dcmitype/StillImage'/>\n"
        "   </cc:Work>\n"
        "   <cc:License rdf:about='http://creativecommons.org/licenses/by/4.0/'>\n"
        "     <cc:permits rdf:resource='http://web.resource.org/cc/Reproduction'/>\n"
        "     <cc:permits rdf:resource='http://web.resource.org/cc/Distribution'/>\n"
        "     <cc:permits rdf:resource='http://web.resource.org/cc/DerivativeWorks'/>\n"
        "     <cc:requires rdf:resource='http://web.resource.org/cc/Notice'/>\n"
        "     <cc:requires rdf:resource='http://web.resource.org/cc/Attribution'/>\n"
        "   </cc:License>\n"
        " </rdf:RDF>\n"
        "</metadata>\n"))
    img.add_line((sympy.Integer(0), sympy.Integer(0)),
                 (sympy.Integer(1), sympy.Integer(1)), symfem.plotting.colors.ORANGE)
    img.save(os.path.join(folder, "test_metadata.svg"))



"""Mardal-Tai-Winther elements on simplices.

This element's definition appears in https://doi.org/10.1137/S0036142901383910
(Mardal, Tai, Winther, 2002)
and https://doi.org/10.1007/s10092-006-0124-6 (Tail, Mardal, 2006)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralMoment, ListOfFunctionals, NormalIntegralMoment,
                           TangentIntegralMoment)
from ..functions import FunctionInput, VectorFunction
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange
from .nedelec import NedelecFirstKind


class MardalTaiWinther(CiarletElement):
    """Mardal-Tai-Winther Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert order == 3

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference, facets=(NormalIntegralMoment, Lagrange, 1,
                               "contravariant", {"variant": variant}))

        poly: typing.List[FunctionInput] = []
        if reference.name == "triangle":
            poly += [(1, 0), (x[0], 0), (x[1], 0),
                     (0, 1), (0, x[0]), (0, x[1]),
                     # (x**2 + 2*x*y, -2*x*y - y**2)
                     (x[0] ** 2 + 2 * x[0] * x[1],
                      -2 * x[0] * x[1] - x[1] ** 2),
                     # (-x**3 + 2*x**2 + 3*x*y**2, 3*x**2*y - 4*x*y - y**3)
                     (-x[0] ** 3 + 2 * x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                      3 * x[0] ** 2 * x[1] - 4 * x[0] * x[1] - x[1] ** 3),
                     # (2*x**2*y + x**2 + 3*x*y**2, -2*x*y**2 - 2*x*y - y**3)
                     (2 * x[0] ** 2 * x[1] + x[0] ** 2 + 3 * x[0] * x[1] ** 2,
                      -2 * x[0] * x[1] ** 2 - 2 * x[0] * x[1] - x[1] ** 3)]
            dofs += make_integral_moment_dofs(
                reference, facets=(TangentIntegralMoment, Lagrange, 0,
                                   "contravariant", {"variant": variant}))
        else:
            assert reference.name == "tetrahedron"

            poly += polynomial_set_vector(reference.tdim, reference.tdim, 1)
            for p in polynomial_set_vector(reference.tdim, reference.tdim, 1):
                poly.append(VectorFunction(tuple(
                    i * x[0] * x[1] * x[2] * (1 - x[0] - x[1] - x[2])
                    for i in p)).curl())

            dofs += make_integral_moment_dofs(
                reference, facets=(IntegralMoment, NedelecFirstKind, 1, "contravariant",
                                   {"variant": variant}))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Mardal-Tai-Winther", "MTW"]
    references = ["triangle", "tetrahedron"]
    min_order = 3
    max_order = 3
    continuity = "H(div)"



"""Create elements and references."""

import importlib as _il
import os as _os
import typing as _typing

from . import references as _references
from .finite_element import FiniteElement as _FiniteElement
from .geometry import SetOfPointsInput as _SetOfPointsInput
from .geometry import parse_set_of_points_input as _parse_set_of_points_input

_folder = _os.path.dirname(_os.path.realpath(__file__))

_elementmap: _typing.Dict[str, _typing.Dict[str, _typing.Type]] = {}
_elementlist: _typing.List[_typing.Type] = []


def add_element(element_class: _typing.Type):
    """Add an element to Symfem.

    Args:
        element_class: The class defining the element.
    """
    global _elementlist
    global _elementmap
    if not isinstance(element_class, type):
        raise TypeError("Element must be defined by a class.")
    if not issubclass(element_class, _FiniteElement):
        raise TypeError("Element must inherit from the FiniteElement class.")
    if element_class == _FiniteElement:
        raise TypeError("Cannot add the FiniteElement class itself.")
    if len(element_class.names) == 0:
        raise TypeError("An element with no names cannot be added")

    if element_class not in _elementlist:
        _elementlist.append(element_class)
    for _n in element_class.names:
        for _r in element_class.references:
            if _n not in _elementmap:
                _elementmap[_n] = {}
            if _r in _elementmap[_n]:
                assert element_class == _elementmap[_n][_r]
            _elementmap[_n][_r] = element_class


for _file in _os.listdir(_os.path.join(_folder, "elements")):
    if _file.endswith(".py") and "__init__" not in _file:
        _fname = _file[:-3]
        _module = _il.import_module(f"symfem.elements.{_fname}")

        for _class_name in dir(_module):
            _element = getattr(_module, _class_name)
            if (
                isinstance(_element, type)
                and issubclass(_element, _FiniteElement)
                and _element != _FiniteElement
                and len(_element.names) > 0
            ):
                add_element(_element)


def create_reference(
    cell_type: str, vertices: _typing.Optional[_SetOfPointsInput] = None
) -> _references.Reference:
    """Make a reference cell.

    Args:
        cell_type: The reference cell type.
                   Supported values: point, interval, triangle, quadrilateral, tetrahedron,
                   hexahedron, prism, pyramid, dual polygon(number_of_triangles)
        vertices: The vertices of the reference.
    """
    args = []
    if vertices is not None:
        args.append(_parse_set_of_points_input(vertices))

    if cell_type == "point":
        return _references.Point(*args)
    if cell_type == "interval":
        return _references.Interval(*args)
    elif cell_type == "triangle":
        return _references.Triangle(*args)
    elif cell_type == "tetrahedron":
        return _references.Tetrahedron(*args)
    elif cell_type == "quadrilateral":
        return _references.Quadrilateral(*args)
    elif cell_type == "hexahedron":
        return _references.Hexahedron(*args)
    elif cell_type == "prism":
        return _references.Prism(*args)
    elif cell_type == "pyramid":
        return _references.Pyramid(*args)
    elif cell_type.startswith("dual polygon"):
        n_tri = int(cell_type.split("(")[1].split(")")[0])
        return _references.DualPolygon(n_tri, *args)
    else:
        raise ValueError(f"Unknown cell type: {cell_type}")


def create_element(
    cell_type: str, element_type: str, order: int, **kwargs: _typing.Any
) -> _FiniteElement:
    """Make a finite element.

    Args:
        cell_type: The reference cell type.
                   Supported values: point, interval, triangle, quadrilateral, tetrahedron,
                   hexahedron, prism, pyramid, dual polygon(number_of_triangles)
        element_type: The type of the element.
                      Supported values:
                      Lagrange, P,
                      vector Lagrange, vP,
                      matrix Lagrange,
                      symmetric matrix Lagrange,
                      dPc,
                      vector dPc,
                      Crouzeix-Raviart, CR, Crouzeix-Falk, CF,
                      conforming Crouzeix-Raviart, conforming CR,
                      serendipity, S,
                      serendipity Hcurl, Scurl, BDMCE, AAE,
                      serendipity Hdiv, Sdiv, BDMCF, AAF,
                      direct serendipity,
                      Regge,
                      Nedelec, Nedelec1, N1curl, Ncurl,
                      Nedelec2, N2curl,
                      Raviart-Thomas, RT, N1div,
                      Brezzi-Douglas-Marini, BDM, N2div,
                      Q,
                      vector Q, vQ,
                      NCE, RTCE, Qcurl,
                      NCF, RTCF, Qdiv,
                      Morley,
                      Morley-Wang-Xu, MWX,
                      Hermite,
                      Mardal-Tai-Winther, MTW,
                      Argyris,
                      bubble,
                      dual polynomial, dual P, dual,
                      Buffa-Christiansen, BC,
                      rotated Buffa-Christiansen, RBC,
                      Brezzi-Douglas-Fortin-Marini, BDFM,
                      Brezzi-Douglas-Duran-Fortin, BDDF,
                      Hellan-Herrmann-Johnson, HHJ,
                      Arnold-Winther, AW, conforming Arnold-Winther,
                      Bell,
                      Kong-Mulder-Veldhuizen, KMV,
                      Bernstein, Bernstein-Bezier,
                      Hsieh-Clough-Tocher, Clough-Tocher, HCT, CT,
                      reduced Hsieh-Clough-Tocher, rHCT,
                      Taylor, discontinuous Taylor,
                      bubble enriched Lagrange,
                      bubble enriched vector Lagrange,
                      Bogner-Fox-Schmit, BFS,
                      Fortin-Soulie, FS,
                      Bernardi-Raugel,
                      Wu-Xu,
                      transition,
                      Guzman-Neilan,
                      nonconforming Arnold-Winther, nonconforming AW,
                      TScurl, trimmed serendipity Hcurl,
                      TSdiv, trimmed serendipity Hdiv,
                      TNT, tiniest tensor,
                      TNTcurl, tiniest tensor Hcurl,
                      TNTdiv, tiniest tensor Hdiv,
                      Arnold-Boffi-Falk, ABF,
                      Arbogast-Correa, AC, AC full, Arbogast-Correa full,
                      Rannacher-Turek,
                      P1-iso-P2, P2-iso-P1, iso-P2 P1,
                      Huang-Zhang, HZ
        order: The order of the element.
    """
    reference = create_reference(cell_type)

    if element_type in _elementmap:
        if reference.name not in _elementmap[element_type]:
            raise ValueError(f"{element_type} element cannot be created on a {reference.name}.")
        element_class = _elementmap[element_type][reference.name]
        if not _order_is_allowed(element_class, reference.name, order):
            raise ValueError(f"Order {order} {element_type} element cannot be created.")
        return element_class(reference, order, **kwargs)

    raise ValueError(f"Unsupported element type: {element_type}")


def _order_is_allowed(
    element_class: _typing.Type, ref: str, order: int
) -> bool:
    """Check that an order is valid for an element.

    Args:
        element_class: The element class
        ref: The reference cell
        order: The polynomial order
    """
    if hasattr(element_class, "min_order"):
        if isinstance(element_class.min_order, dict):
            if ref in element_class.min_order:
                if order < element_class.min_order[ref]:
                    return False
        elif order < element_class.min_order:
            return False
    if hasattr(element_class, "max_order"):
        if isinstance(element_class.max_order, dict):
            if ref in element_class.max_order:
                if order > element_class.max_order[ref]:
                    return False
        elif order > element_class.max_order:
            return False
    return True



"""Test DOF descriptions."""

import pytest

from symfem import create_element
from symfem.finite_element import CiarletElement

from .utils import test_elements


@pytest.mark.parametrize(
    ("cell_type", "element_type", "order", "kwargs"),
    [[reference, element, order, kwargs]
     for reference, i in test_elements.items() for element, j in i.items()
     for kwargs, k in j for order in k])
def test_element(
    elements_to_test, cells_to_test, cell_type, element_type, order, kwargs,
    speed
):
    """Run tests for each element."""
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell_type not in cells_to_test:
        pytest.skip()
    if speed == "fast":
        if order > 2:
            pytest.skip()
        if order == 2 and cell_type in ["tetrahedron", "hexahedron", "prism", "pyramid"]:
            pytest.skip()

    element = create_element(cell_type, element_type, order, **kwargs)
    if isinstance(element, CiarletElement):
        for d in element.dofs:
            print(d.get_tex()[0])
